import { BasicWebSocket } from "samsara-common/graphql/basicwebsocket";
import { PingingWebSocket } from "samsara-common/graphql/pingingwebsocket";

import {
  addPings,
  newMockListener,
  newMockWebSocket,
} from "samsara-common/graphql/mocks/websocket";

jest.useFakeTimers();

test("connect never returns websocket", async () => {
  const listener = newMockListener();

  const pws = new PingingWebSocket(
    () => new Promise<BasicWebSocket>(() => undefined),
    listener,
  );
  pws.connect();

  jest.runTimersToTime(60000);

  await listener.calls([{ close: PingingWebSocket.reasonConnectionTimeout }]);
  listener.done();
});

test("connect fails", async () => {
  const listener = newMockListener();

  const pws = new PingingWebSocket(
    () => Promise.reject(new Error("bad time")),
    listener,
  );
  pws.connect();

  await listener.calls([{ close: "connectFunction: Error: bad time" }]);
  listener.done();
});

test("websocket never opens", async () => {
  const listener = newMockListener();

  const ws = newMockWebSocket();
  const pws = new PingingWebSocket(async () => ws, listener);
  await pws.connect();

  jest.runTimersToTime(60000);

  await listener.calls([{ close: PingingWebSocket.reasonConnectionTimeout }]);
  listener.done();
});

test("websocket opens but does not ping", async () => {
  const listener = newMockListener();

  const ws = newMockWebSocket();
  const pws = new PingingWebSocket(async () => ws, listener);
  await pws.connect();

  ws.readyState = BasicWebSocket.OPEN;
  ws.onopen(new Event("open"));

  jest.runTimersToTime(60000);

  await listener.calls([
    { open: true },
    { close: PingingWebSocket.reasonPingTimeout },
  ]);
  listener.done();
});

test("websocket opens and pings", async () => {
  const listener = newMockListener();

  const ws = newMockWebSocket();
  const pws = new PingingWebSocket(async () => ws, listener);
  await pws.connect();

  addPings(ws);

  ws.readyState = BasicWebSocket.OPEN;
  ws.onopen(new Event("open"));

  await listener.calls([{ open: true }]);

  jest.runTimersToTime(60000);

  await listener.calls([
    { message: { type: "echo" } },
    { message: { type: "echo" } },
  ]);

  pws.close();

  await listener.calls([{ close: PingingWebSocket.reasonCloseCalled }]);
  listener.done();
});

test("websocket closes", async () => {
  const listener = newMockListener();

  const ws = newMockWebSocket();
  const pws = new PingingWebSocket(async () => ws, listener);
  await pws.connect();

  addPings(ws);

  ws.readyState = BasicWebSocket.OPEN;
  ws.onopen(new Event("open"));

  await listener.calls([{ open: true }]);

  jest.runTimersToTime(30000);

  await listener.calls([{ message: { type: "echo" } }]);

  ws.readyState = BasicWebSocket.CLOSED;
  ws.onclose({ reason: "reason" } as CloseEvent);

  await listener.calls([{ close: "socket close: reason" }]);
  listener.done();
});

test("websocket close throws", async () => {
  const listener = newMockListener();

  const ws = newMockWebSocket();
  const pws = new PingingWebSocket(async () => ws, listener);
  await pws.connect();

  addPings(ws);

  ws.readyState = BasicWebSocket.OPEN;
  ws.onopen(new Event("open"));

  await listener.calls([{ open: true }]);

  ws.close = () => {
    throw new Error("boo!");
  };

  pws.close();

  await listener.calls([{ close: PingingWebSocket.reasonCloseCalled }]);
  listener.done();
});

test("websocket send throws", async () => {
  const listener = newMockListener();

  const ws = newMockWebSocket();
  const pws = new PingingWebSocket(async () => ws, listener);
  await pws.connect();

  ws.readyState = BasicWebSocket.OPEN;
  ws.onopen(new Event("open"));

  await listener.calls([{ open: true }]);

  ws.send = () => {
    throw new Error("yikes!");
  };

  pws.send({ type: "echo" });

  pws.close();

  await listener.calls([{ close: PingingWebSocket.reasonCloseCalled }]);
  listener.done();
});

test("don't send on non-open websocket", async () => {
  const listener = newMockListener();

  const ws = newMockWebSocket();
  const pws = new PingingWebSocket(async () => ws, listener);
  await pws.connect();

  ws.readyState = BasicWebSocket.CLOSED;

  let sendCalled = false;
  ws.send = () => {
    sendCalled = true;
    throw new Error("not connected");
  };

  pws.send({ type: "echo" });
  expect(sendCalled).toBe(false);

  pws.close();
  await listener.calls([{ close: PingingWebSocket.reasonCloseCalled }]);
  listener.done();
});

test("websocket closes after invalid json", async () => {
  const listener = newMockListener();

  const ws = newMockWebSocket();
  const pws = new PingingWebSocket(async () => ws, listener);
  await pws.connect();

  ws.readyState = BasicWebSocket.OPEN;
  ws.onopen(new Event("open"));

  await listener.calls([{ open: true }]);

  ws.onmessage(new MessageEvent("message", { data: `{"foo": "bar"}` }));

  await listener.calls([{ message: { foo: "bar" } }]);

  ws.onmessage(new MessageEvent("message", { data: `i am not json!!!` }));

  await listener.calls([
    {
      close:
        "socket message JSON.parse: SyntaxError: Unexpected token i in JSON at position 0",
    },
  ]);
  listener.done();
});
