import { BasicWebSocket } from "samsara-common/graphql/basicwebsocket";
import { PingingWebSocket } from "samsara-common/graphql/pingingwebsocket";
import {
  reconnectDelayMs,
  ReconnectingWebSocket,
} from "samsara-common/graphql/reconnectingwebsocket";

import {
  addPings,
  MockWebSocket,
  newMockListener,
  newMockWebSocket,
} from "samsara-common/graphql/mocks/websocket";

jest.useFakeTimers();

async function runForMs(d: number) {
  jest.runTimersToTime(d);
  return Promise.resolve();
}

test("immediate reconnect after successful pings", async () => {
  const listener = newMockListener();

  let ws = newMockWebSocket();
  addPings(ws);

  const rws = new ReconnectingWebSocket(async () => ws, listener);

  // Open the connection.
  await rws.reconnect();

  // Fire the connect event.
  ws.readyState = BasicWebSocket.OPEN;
  ws.onopen(new Event("open"));
  // Expect the connect event callback.
  await listener.calls([{ open: true }]);

  // Wait for two pings.
  await runForMs(PingingWebSocket.defaultOptions.pingIntervalMs * 2);
  await listener.calls([
    { message: { type: "echo" } },
    { message: { type: "echo" } },
  ]);

  // Fire a close event.
  ws.readyState = BasicWebSocket.CLOSED;
  ws.onclose({ reason: "bye" } as CloseEvent);
  // Expect the close event callback.
  await listener.calls([{ close: "socket close: bye" }]);

  // Prepare a new websocket for the next connection.
  ws = newMockWebSocket();
  addPings(ws);
  // Run timers to let the new websocket be picked up.
  await runForMs(0);

  // Fire the connect.
  ws.readyState = BasicWebSocket.OPEN;
  ws.onopen(new Event("open"));
  // And expect the callback.
  await listener.calls([{ open: true }]);

  // Wait for two pings.
  runForMs(PingingWebSocket.defaultOptions.pingIntervalMs * 2);
  await listener.calls([
    { message: { type: "echo" } },
    { message: { type: "echo" } },
  ]);

  // Close the connection.
  rws.close();
  await listener.calls([{ close: "close called" }]);

  listener.done();
});

test("delayed reconnect with no successful pings", async () => {
  const listener = newMockListener();

  let ws: MockWebSocket | Promise<never> = newMockWebSocket();

  const rws = new ReconnectingWebSocket(async () => ws, listener);

  // Open the connection.
  await rws.reconnect();

  // Fire the connect event.
  ws.readyState = BasicWebSocket.OPEN;
  ws.onopen(new Event("open"));
  // Expect the connect callback.
  await listener.calls([{ open: true }]);

  // Run for a while, but not quite the ping interval.
  await runForMs(PingingWebSocket.defaultOptions.pingIntervalMs - 500);
  // Fire a close event.
  ws.readyState = BasicWebSocket.CLOSED;
  ws.onclose({ reason: "bye" } as CloseEvent);
  await listener.calls([{ close: "socket close: bye" }]);

  // Set ws to reject to fail future connects, and run for not quite the
  // reconnect delay to verify that we don't connect yet.
  ws = Promise.reject(new Error("failure"));
  await runForMs(reconnectDelayMs - 500);
  ws.catch(() => {
    return;
  });

  // Now prepare the next connection.
  ws = newMockWebSocket();
  addPings(ws);
  // Run until the reconnect delay.
  await runForMs(500);

  // Fire the connect event.
  ws.readyState = BasicWebSocket.OPEN;
  ws.onopen(new Event("open"));
  // And verify the callback.
  await listener.calls([{ open: true }]);

  // See pings.
  await runForMs(PingingWebSocket.defaultOptions.pingIntervalMs * 2);
  await listener.calls([
    { message: { type: "echo" } },
    { message: { type: "echo" } },
  ]);

  // Close the connection.
  rws.close();
  await listener.calls([{ close: "close called" }]);

  listener.done();
});

test("delayed reconnect with connect timeouts", async () => {
  const listener = newMockListener();

  let ws: MockWebSocket | Promise<never> = newMockWebSocket();

  const rws = new ReconnectingWebSocket(async () => ws, listener);

  // Open the connection.
  await rws.reconnect();

  // See repeated connect attempts, evenly spaced.
  for (let i = 0; i < 10; i++) {
    // Expect no reconnects during the connection timeout and reconnect delay.
    ws = Promise.reject(new Error("failure"));
    // Wait for the connection timeout and expect a timeout.
    await runForMs(PingingWebSocket.defaultOptions.connectionTimeoutMs);
    await listener.calls([{ close: PingingWebSocket.reasonConnectionTimeout }]);
    // Wait for the reconnect delay.
    await runForMs(reconnectDelayMs - 500);

    ws.catch(() => {
      return;
    });

    // Now allow a new connection.
    ws = newMockWebSocket();
    await runForMs(500);
  }

  // Close the connection.
  rws.close();
  await listener.calls([{ close: PingingWebSocket.reasonCloseCalled }]);

  listener.done();
});

test("no reconnects after close on successfully opened connection", async () => {
  const listener = newMockListener();

  let ws: MockWebSocket | Promise<never> = newMockWebSocket();
  addPings(ws);

  const rws = new ReconnectingWebSocket(async () => ws, listener);

  // Open the connection.
  await rws.reconnect();

  // Fire the open event.
  ws.readyState = BasicWebSocket.OPEN;
  ws.onopen(new Event("open"));
  // Expect a callback.
  await listener.calls([{ open: true }]);

  // Expect pings.
  await runForMs(PingingWebSocket.defaultOptions.pingIntervalMs * 2);
  await listener.calls([
    { message: { type: "echo" } },
    { message: { type: "echo" } },
  ]);

  // Close the connection and fail reconnects.
  ws = Promise.reject(new Error("failure"));
  rws.close();

  // Expect a close callback.
  await listener.calls([{ close: PingingWebSocket.reasonCloseCalled }]);

  // Run for a while without connect being called.
  await runForMs(
    10 *
      (reconnectDelayMs + PingingWebSocket.defaultOptions.connectionTimeoutMs),
  );

  // Prepare a new connection and allow a reconnection.
  ws = newMockWebSocket();
  addPings(ws);
  await rws.reconnect();

  // Fire the open event.
  ws.readyState = BasicWebSocket.OPEN;
  ws.onopen(new Event("open"));
  // Expect a callback.
  await listener.calls([{ open: true }]);

  // Close the connection.
  rws.close();
  await listener.calls([{ close: PingingWebSocket.reasonCloseCalled }]);

  listener.done();
});

test("no reconnects after close during reconnect wait", async () => {
  const listener = newMockListener();

  let ws: MockWebSocket | Promise<never> = newMockWebSocket();
  addPings(ws);

  const rws = new ReconnectingWebSocket(async () => ws, listener);

  // Open the connection.
  await rws.reconnect();

  // Fail future connections.
  ws = Promise.reject(new Error("failure"));

  // Wait for the connection timeout to fire.
  await runForMs(PingingWebSocket.defaultOptions.connectionTimeoutMs);
  await listener.calls([{ close: PingingWebSocket.reasonConnectionTimeout }]);

  // Wait not quite the reconnection delay.
  await runForMs(reconnectDelayMs - 500);

  // Call close to prevent future connections.
  rws.close();

  // Verify there are no connections (if there were, they would
  // fail with the above error.)
  await runForMs(
    10 *
      (reconnectDelayMs + PingingWebSocket.defaultOptions.connectionTimeoutMs),
  );

  // Prepare a new connection.
  ws = newMockWebSocket();
  addPings(ws);

  // Reopen the socket.
  await rws.reconnect();

  // Fire the open event.
  ws.readyState = BasicWebSocket.OPEN;
  ws.onopen(new Event("open"));
  // Expect a callback.
  await listener.calls([{ open: true }]);

  // Close the connection.
  rws.close();
  await listener.calls([{ close: PingingWebSocket.reasonCloseCalled }]);

  listener.done();
});
